library(glmnet)
library(dplyr)
library(data.table)
library(parallel)
library(selectiveInference)
library(AUC)
library(MASS)

setwd(Sys.getenv('PCS_DIR'))
load('data/enhancer.Rdata')
source('scripts/simulation_utilities.R')

ncore <- 48
set.seed(47)
test.prop <- 0.5 # proportion test data
s.g <- 1 # scaling factor for number of active features
n.g <- c(250, 1000) # number of observations
snr.g <- 2 # signal to noise ratio
p.g <- 630 # number of features
fam.g <- c('gaussian') # family for responses
beta.g <- c('fixed') # beta fixed or decaying
design.g <- c('tf') # design matrix
nbs <- 100 # number of bootstrap samples for stability analysis
nsim <- 48 # number of replications
params <- expand.grid(s.g, n.g, snr.g, p.g, 
                      fam.g, beta.g, design.g, 
                      simplify=FALSE) 

# Expand data to include polynomial terms
pairs <- combn(ncol(x), 2, simplify=FALSE)
xpair <- sapply(pairs, function(i) x[,i[1]] * x[,i[2]])
xpair <- scale(xpair)
xpair <- xpair + rnorm(nrow(xpair) * ncol(xpair))
x <- cbind(x, xpair)
x <- scale(x)
colnames(x) <- paste0('X', 1:ncol(x))
p <- ncol(x)

for (i in 1:nrow(params)) {

  n <- params[i, 2]
  snr <- params[i, 3]
  p <- params[i, 4]
  s <- floor(params[i, 1] * sqrt(p))
  design <- params[i, 7]
  beta <- params[i, 6]
  family <- as.character(params[i, 5])

  # Generate covaraiance matrices based on data and block structure
  cmat <- blockCov(n, k=100, rho=0.8)
  param.str <- paste0('s-', s, '_n-', n, '_snr-', snr, '_p-', p, 
                      '_x-', design, '_beta-', beta, '_f-', family)
  pdir <- paste0('results_cv_final/sims/')
  dir.create(pdir, recursive=TRUE)
  cat(paste0('Sim: ', param.str, '...\n'))
  
  out <- mclapply(1:nsim, function(ii) {
    
    set.seed(ii)
    cat(paste('param: ', i, '--', 'rep: ', ii, '...\n'))
    # Sample n observations for simulation  
    ss.obs <- sample(nrow(x), n) 
    if (design == 'tf') {
      xx <- x[ss.obs, sample(ncol(x), p)]
      xx <- scale(xx)
    } else {
      xx <- matrix(rnorm(n * p), nrow=n)
      varnames <- paste0('X', 1:ncol(xx))
      colnames(xx) <- varnames
      xx <- scale(xx)
    }

    varnames <- colnames(xx)
    train.id <- sample(n, n * (1 - test.prop))
    test.id <- setdiff(1:n, train.id)

    # Generate resposnes as sparse linear model
    betas <- rep(0, p)
    if (s > 0) {
      beta.id <- sample(1:p, s)
      beta.rm <- sample(beta.id, floor(s / 2))
      if (beta == 'fixed') {
        betas[beta.id] <- 1 / sqrt(s)
      } else {
        betas[beta.id] <- rexp(s, sqrt(s))
      }
      ytrue <- xx %*% betas 
      yls3 <- ls3(xx, betas)
    } else {
      beta.id <- NULL
      ytrue <- rep(0, n)
    }
    
    # Generate error terms for each model
    err <- list()
    err$gauss <- rnorm(n)
    err$t <- rt(n, df=3)
    err$b100 <- mvrnorm(n=1, mu=rep(0, n), Sigma=cmat)
    xscale <- sqrt(rowMeans(xx ^ 2))
    err$dep <- sapply(xscale, rnorm, n=1, mean=0)

    # Adjust noise for specified SNR
    if (s > 0) {
      y <- addNoise(ytrue, err, snr)
    } else {
      y <- err 
    }

    ## Datasets to be used for simulation
    simdata <- lapply(y, function(yy) {
                        out <- list()
                        out$x <- xx
                        out$y <- yy
                        return(out)
    })

    if (s > 0) {
      nn <- length(y)
      simdata[[nn + 1]] <- simdata[[1]]
      simdata[[nn + 1]]$x <- simdata[[nn + 1]]$x[,-beta.rm]
      simdata[[nn + 2]] <- simdata[[1]]
      simdata[[nn + 2]]$y <- unlist(addNoise(yls3, err[1], snr))
    }
    
    # Run inference on simualted data
    pcs <- lapply(simdata, function(z) {
      fitPCSWrap(x=z$x, y=z$y, train.id=train.id, test.id=test.id, 
                 family=family, nbs=nbs)
    })

    cv <- ifelse(s == 0, FALSE, TRUE)
    selinf <- lapply(simdata, function(z) {
        fitSelInf(x=z$x, y=z$y, family=family, cv=cv)
    })

    flm <- lapply(simdata, function(z) {
      fitLM(x=z$x, y=z$y, family=family)
    })

    param <- c('Gaussian', 't_3', 'Block', 'Dependent', 
               'Misspecified_miss', 'Misspecified_rule')
    out <- aggregateInf(pcs, selinf, flm, param, beta.id, varnames) %>% 
      mutate(rep=ii)
    
    return(out)
  }, mc.cores = ncore)

  id.rm <- sapply(out, is.null)
  out <- rbindlist(out[!id.rm])
  save(file=paste0(pdir,   param.str, '.Rdata'), out)
}

