## Inference functions
fitPCSWrap <- function(x, y, train.id, test.id, nbs, k=10, family='gaussian') {
  ## Wrapper function to run PCS inference across bootstrap replicates
  id <- c(train.id, test.id)
  bsid <- replicate(nbs, bootstrapSample(id), simplify=FALSE)
  
  # select lambda from each sample split
  lam1 <- selectLambda(x, y, train.id, test.id, family, k)
  lam2 <- selectLambda(x, y, test.id, train.id, family, k)
 
  # run PCS inference across bootstrap samples 
  out <- sapply(bsid, function(i) {
    tr.id <- i[i %in% train.id]
    tst.id <- i[i %in% test.id]
    f1 <- fitPCS(x=x, y=y, train.id=tr.id, test.id=tst.id, lam=lam1, family=family)
    f2 <- fitPCS(x=x, y=y, train.id=tst.id, test.id=tr.id, lam=lam2, family=family)
    return((f1 + f2) / 2)
  })
  
  return(1 - rowMeans(out))
}

fitPCS <- function(x, y, train.id, test.id, lam, family='gaussian') {
  ## Conduct PCS iinference using penealty parameter model perturbations.
  require(glmnet)
  require(AUC)
  fit <- glmnet(x=x[train.id,], y=y[train.id], lambda=lam,
                family=family, standardize=FALSE)
  beta.prop <- rowMeans(as.matrix(fit$beta) != 0)
  return(beta.prop)
}

selectLambda <- function(x, y, train.id, test.id, family, k) {
  # Select a set of optimal penalty parameters using held out data
  fit <- glmnet(x=x[train.id,], y=y[train.id],
                family=family, standardize=FALSE)
  yhat <- predict(fit, newx=x[test.id,], type='response')

  if (family == 'gaussian') {
    err <- colMeans((yhat - y[test.id]) ^ 2)
  } else {
    ytst <- as.factor(y[test.id])
    err <- apply(yhat, MAR=2, function(z) auc(roc(predictions=z, labels=ytst)))
    err <- 1 - err
  }
  
  err.min <- order(err)[1:k]
  lambdas <- fit$lambda
  lsel <- lambdas %in% lambdas[err.min]
  return(lambdas[lsel])
}

fitSelInf <- function(x, y, cv=TRUE, family='gaussian') {
  # Conduct selective inference for linear model. If cv = TRUE, model will be
  # selected  using cross validation, otherwise model will be selected to
  # achieve a model size of 2 * sqrt(p)
  require(selectiveInference)
  require(glmnet)
  n <- nrow(x)

  # select model using cross validation or minimum size requirement
  if (cv) {
    fit <- cv.glmnet(x=x, y=y, family=family, nfolds=5, thresh=1e-12, standardize=FALSE)
    lambda <- fit$lambda.min
  } else {
    fit <- glmnet(x=x, y=y, family=family, thresh=1e-12, standardize=FALSE)
    idsel <- which(colSums(fit$beta != 0) > 2 * sqrt(ncol(x)))
    if (length(idsel) == 0) idsel <- length(fit$lambda)
    if (length(idsel) > 1) idsel <- min(idsel)
    lambda <- fit$lambda[idsel]
  }

  # extract predicted responses and parameter
  yhat <- predict(fit, newx=x, s=lambda)
  beta <- coef(fit, x=x, y=y, s=lambda, exact=TRUE)

  if (family == 'gaussian') {
    beta <- beta[-1]
    nz <- sum(beta != 0)
    sigma <- sqrt(sum((y - yhat)^2)/(length(y) - nz - 1))
  } else {
    sigma <- NULL
  }

  # evaluate p-values using selective inference
  pval <- rep(1, ncol(x))
  out <- fixedLassoInf(x, y, beta, lambda * n, family=family, sigma=sigma)
  pval[out$vars] <- out$pv
  names(pval) <- colnames(x)
  return(pval)
}

fitLM <- function(x, y, prop=0.2, family='gaussian') {
  # Conduct inferences using linear model t-test after lasso selection
  require(glmnet)

  # select active features with lasso CV
  fit <- cv.glmnet(x=x, y=y, family=family, nfolds=5, standardize=FALSE)
  lambda <- fit$lambda.min
  beta.id <- (coef(fit, x=x, y=y, s=lambda, exact=TRUE) != 0)[-1]

  # fit linear model using selected features and conduct inference
  d <- data.frame(x=x[,beta.id], y=y)
  if (family == 'gaussian')
    fit <- lm(y ~ ., data=d)
  else
    fit <- glm(y ~ ., data=d, family='binomial')

  pval <- rep(1, ncol(x))
  s <- summary(fit)
  pval.coef <- s$coefficients[-1,4]
  pval[beta.id] <- pval.coef
  names(pval) <- colnames(x)
  return(pval)
}

aggregateInf <- function(pcsinf, selinf, lminf, param, beta.id, varnames) {
  require(data.table)
  out <- mapply(function(rpcs, rsel, rlm, par) {
    mfpcs <- melt(rpcs) %>% mutate(Var1=names(rpcs),
                                   active=(Var1 %in% varnames[beta.id]),
                                   noise=par,
                                   inference='PCS')
    mfsel <- melt(rsel) %>% mutate(Var1=names(rsel),
                                   active=(Var1 %in% varnames[beta.id]),
                                   noise=par,
                                   inference='Sel_Inf')
    mflm <- melt(rlm) %>% mutate(Var1=names(rlm),
                                 active=(Var1 %in% varnames[beta.id]),
                                 noise=par,
                                 inference='Linear_Model')
    return(rbind(mfpcs, mfsel, mflm))
  }, pcsinf, selinf, lminf, param, SIMPLIFY = FALSE)
  out <- rbindlist(out)
}

## Data generating functions
bootstrapSample <- function(id) {
  bsid <- sample(id, length(id), replace=TRUE)
  return(bsid)
}

ls3 <- function(x, betas, k=2) {
  # Generate responses as a linear combination of order-k rules
  beta.act <- which(betas != 0)
  ngrp <- length(beta.act) / k
  grp <- rep(1:floor(ngrp), times=k)
  if (floor(ngrp) != ngrp) grp <- c(grp, ceiling(ngrp))
  grp <- sample(grp)

  ints <- split(beta.act, grp)
  medThresh <- function(z) z >= median(z[z != 0])
  intRule <- function(z) rowMeans(apply(as.matrix(z), MAR=2, medThresh)) == 1
  yy <- sapply(ints, function(z) as.numeric(intRule(x[,z])) * sum(betas[z]))
  return(rowSums(yy))

}

blockCov <- function(n, k, rho) {
  ## Generate block covariance with k blocks and constant off diagonal rho
  out.mat <- matrix(0, nrow=n, ncol=n)
  block.size <- floor(n / k)
  for (kk in 1:k) {
    id <- ((kk - 1) * block.size + 1):(kk * block.size)
    out.mat[id, id] <- rho
  }
  diag(out.mat) <- 1
  return(out.mat)
} 

addNoise <- function(ytrue, error, snr) {
  # Add noise to achieve specified signal to noise ratio
  k <- c(sqrt(var(ytrue) / snr))
  y <- lapply(error, function(e) ytrue + k * scale(e))  
  return(y)
}

## Plotting functions for feature selection
getCurves <- function(x, par.grid, group=TRUE, grid=seq(0, 1, by=1e-2), ncore=1) {
  # Wraper function to evaluate ROC curve across multiple parameter settings
  require(parallel)
  require(data.table)

  curves <- mcmapply(function(pp, ii, rr) {
    cc <- getCurve(x, rr, ii, pp, grid=grid)
    out <- data.frame(fpr=grid,
                      tpr=cc,
                      rep=rr,
                      noise=pp,
                      inference=ii)
    return(out)
  }, param.grid$Var1, param.grid$Var2, param.grid$Var3, 
  mc.cores=ncore, SIMPLIFY=FALSE)
  
  curves <- rbindlist(curves) 
  
  if (group) {
  curves <- curves %>%
    group_by(noise, inference, fpr) %>%
    summarise(tpr=mean(tpr))
  }
  return(curves)
}

getCurve <- function(x, rep.sel, inf.sel, noise.sel, grid=seq(0, 1, by=1e-2)) {
  # Evaluate the ROC curve for a given replicate and given parameters
  require(dplyr)
  require(AUC)
  xf <- filter(x, rep == rep.sel, inference == inf.sel, noise == noise.sel)
  xcurve <-  roc(predictions=1 - xf$value, labels = as.factor(xf$active))
  tpr <- xcurve$tpr
  fpr <- xcurve$fpr
  tpr.grid <- sapply(grid, function(g) max(tpr[fpr <= g]))
  tpr.grid[1] <- 0
  return(tpr.grid)
}
