compareInts <- function(x, file) {
  # loads interaction data from separate analysis for comparison
  if (!file.exists(file)) return(rep(TRUE, nrow(ints)))

  load(file)
  return(x$int %in% stability$int)
}

aucPR <- function(ypred, ytest) {
  require(PRROC)
  pr <- pr.curve(scores.class0=ypred, weights.class0=ytest)
  return(pr$auc.integral)
}

blockbs <- function(x) {
  # generate block bootstrap from list of blocks
  x <- unlist(sample(x, length(x), replace=TRUE))
  return(x)
}

bssample <- function(n) {
  # generate bootstrap sample of n indices
  return(sample(n, n, replace=TRUE))
}

testBlock <- function(n, test.prop=0.25) { 
  # block training test set to either first 25% or last 25% of a chromosome
  dir <- sample(0:1, 1)

  ntest <- ceiling(n * test.prop)
  if (dir == 0)  id <- rep(2:1, times=c(ntest, n - ntest))
  else id <- rep(1:2, times=c(n - ntest, ntest))
  return(c('train', 'test')[id])
}

makeBlocks <- function(gene.coords, idcs, size=5) {
  # Wrapper function to generate blocks for bootstrap sampling by chromosome
  gene.coords <- filter(gene.coords, 1:n() %in% idcs) %>%
    mutate(idx=1:n()) %>%
    group_by(chr) %>%
    arrange(desc(r6start)) %>%
    mutate(idx.chr=1:n()) %>%
    ungroup()
  
  chrBlocksSize <- function(x) chrBlocks(x, size=size)
  block <- c(by(gene.coords, gene.coords$chr, chrBlocksSize))
  block <- unlist(block, recursive=FALSE)
  return(block)
}

chrBlocks <- function(gene.coords, size=5) {
  # Generate blocks for bootstrap sampling for a single chromosome
  idx.breaks <- seq(0, max(gene.coords$idx.chr) + size, by=size)
  idx.cut <- as.numeric(cut(gene.coords$idx.chr, idx.breaks))
  idx.block <- lapply(unique(idx.cut), function(i) 
    gene.coords$idx[idx.cut == i])
  return(idx.block)
}

intSubsets <- function(int) {
  # Generate all order s - 1 subsets of an order s interaction
  require(stringr)
  
  int <- as.character(int)
  int <- unlist(str_split(int, '_'))
  sub.ord <- length(int) - 1
  subs <- lapply(sub.ord, combn, x=int, simplify=FALSE)
  subs <- unlist(subs, recursive=FALSE)
  subs <- sapply(subs, paste, collapse='_')
  return(subs)
}

evaluatePredictions <- function(int, bsaucs) {
  # Compare prediction accuracy of an interaction to lower order subsets
  int.subsets <- intSubsets(int)
  bs.subsets <- as.matrix(bsaucs[,int.subsets])
  auc.subs <- colMeans(bs.subsets)
  id.select <- names(which.max(auc.subs))
  prop.greater <- mean(bsaucs[,id.select] >= bsaucs[,int])
  diff <- mean(bsaucs[,int]) - mean(bsaucs[,id.select])
  return(c(p=prop.greater, diff=diff))
}

bootstrapAUC <- function(bsid, y, predictions, test.id) {
  # Evaluate AUC across bootstrap samples of test set
  pred <- lapply(predictions, '[', bsid)
  aucs <- sapply(pred, aucPR, ytest=y[test.id[bsid]])
  return(aucs)
}

aucPR <- function(ypred, ytest) {
  # Wrapper function for calculating AUC-PR
  require(PRROC)
  pr <-  pr.curve(scores.class0=ypred, weights.class0=ytest)$auc.integral
  return(pr)
}

testBlock <- function(n, test.prop=0.5) {
  # block training test set to either first 50% or last 5% of a sequence
  dir <- sample(0:1, 1)
  
  ntest <- ceiling(n * test.prop)
  if (dir == 0)  id <- rep(1:2, times=c(ntest, n - ntest))
  else id <- rep(2:1, times=c(n - ntest, ntest))
  return(c('train', 'test')[id])
}

normalize <- function(z) z / sqrt(sum(z ^ 2))

plotStab <- function(x, z=NULL, leg=TRUE) {
  p <- x %>%
    transmute(TF=Var1, proportion=value, noise=noise, active=active) %>%
    ggplot() + 
    geom_boxplot(aes(x=TF, y=proportion, fill=active, col=active)) +
    facet_wrap(~noise) +
    theme_bw() + 
    theme(axis.text.x = element_text(angle = 90, hjust = 1))
  
  d <- ggplot_build(p)$data[[1]]
  p <- p + geom_segment(data=d, aes(x=xmin, xend=xmax, y=middle, yend=middle))
  if (!is.null(z)) {
    cols <- ifelse(z$value < 10e-5, '#961623', 'black')
    p <- p + geom_point(data=z, aes(x=Var1, y=(1 - value)),
                        shape=8, col=cols)
  }
  if (!leg) p <- p + guides(fill=FALSE, col=FALSE)
  plot(p)
}
